
WITH Compression_Support;
WITH Compression_Instance;
WITH Direct_Memory_Action;

PROCEDURE Compression_Decode IS

  -- This procedure demonstrates how to use the compression algorithm
  -- to decompress a section of memory.
  --
  -- NOTE: The Direct_Memory_Action package is used to simulate the
  --       operation of the algorithm directly in memory by specifying
  --       the physical address of the buffers to read and write from.
  --
  --       That is why the package Sample_Memory is using the 'ADDRESS
  --       clause to return an actual physical address.  This is not 
  --       something that should normally be done as it it not necessary,
  --       but is used here for illustrative purposes only.
 


  PACKAGE Sample_Memory IS

      Source_List : Compression_Instance.Working_Space;
      Target_List : Compression_Instance.Working_Space;

      Source_Head : Compression_Instance.Object_Index := 0;
      Source_Tail : Compression_Instance.Object_Index := 0;

      Target_Head : Compression_Instance.Object_Index := 0;
      Target_Tail : Compression_Instance.Object_Index := 0;

      Source_Address : CONSTANT Direct_Memory_Action.Memory_Index :=
                       Direct_Memory_Action.Convert( Source_List'ADDRESS );

      Target_Address : CONSTANT Direct_Memory_Action.Memory_Index :=
                       Direct_Memory_Action.Convert( Target_List'ADDRESS );

  END Sample_Memory;



  PROCEDURE Perform_Uncompress IS NEW Direct_Memory_Action.Perform_Action
    ( Object_Value => Compression_Instance.Object_Value,
      Object_Index => Compression_Instance.Object_Index,
      Object_Array => Compression_Instance.Object_Array,
      Operate_On   => Compression_Instance.Uncompress_Data );


BEGIN

  -- Read compression parameters and initialize package states.
  Compression_Support.Initialize_Compression( "Compression Decode Utility" );

  -- Read compressed memory image...
  Compression_Support.Load_Memory
    ( Sample_Memory.Source_List,
      Sample_Memory.Source_Head,
      Sample_Memory.Source_Tail );

  -- Now Source_List contains:
  --  ( Source_List'FIRST .. Source_Tail )       => Entire compressed image
  --  ( Source_List'FIRST .. Source_Head - 1 )   => Do not decompress
  --  ( Source_Head .. Source_Tail )             => To be decompressed

  Compression_Support.Print( "Uncompressing memory map file..." );

  Compression_Support.Start_Timer;

  Perform_Uncompress
    ( Source_Address   => Sample_Memory.Source_Address,
      Source_List_Head => Sample_Memory.Source_List'FIRST,
      Source_List_Tail => Sample_Memory.Source_Tail,
      Source_Head      => Sample_Memory.Source_Head,
      Target_Address   => Sample_Memory.Target_Address,
      Target_List_Head => Sample_Memory.Target_List'FIRST,
      Target_List_Tail => Sample_Memory.Target_List'LAST,
      Target_Tail      => Sample_Memory.Target_Tail );

  Compression_Support.Stop_Timer;

  -- Now Target_List contains:
  --  ( Target_List'FIRST .. Target_Tail )       => Entire uncompressed image
  --  ( Target_List'FIRST .. Source_Head - 1 )   => Same as Source_List
  --  ( Source_Head .. Target_Tail )             => Uncompressed portion
 
  Sample_Memory.Target_Head := Sample_Memory.Source_Head;

  -- Save uncompressed memory image...
  Compression_Support.Save_Memory
    ( Sample_Memory.Target_List,
      Sample_Memory.Target_Head,
      Sample_Memory.Target_Tail );

  Compression_Support.Print_Report
    ( "Uncompress:",
      Sample_Memory.Source_List,
      Sample_Memory.Source_Head,
      Sample_Memory.Source_Tail,
      Sample_Memory.Target_List,
      Sample_Memory.Target_Head,
      Sample_Memory.Target_Tail );

END Compression_Decode;
