
WITH Compression_Primitives;

PACKAGE BODY Compression_Instance IS

  -- Compression arguments:

  TYPE Simple_Count IS NEW Object_Count; -- To make operators visible.
  TYPE Count_Values IS ARRAY( Compression_Parameters.Length_Parameter )
                    OF Simple_Count;


  -- Marker token definition:

  TYPE Stored_Index IS RANGE 1 .. 3;
  TYPE Stored_Array IS ARRAY( Stored_Index ) OF Object_Value;


  -- State data needed by the token conversion routines.  This data can
  -- be changed as needed for different implementations of the functions
  -- Has_A_Marker, Change_Token, Length_Value, and Backup_Value.

  TYPE Change_State IS RECORD
      Totals_Count : Simple_Count := 0;
      Length_Count : Simple_Count := 0;
      Cutoff_Count : Simple_Count := 0;
      Marker_Entry : Compression_Parameters.Marker_Items :=
                     ( Size => 1, List => ( 1 => 0 ) );
  END RECORD;



  FUNCTION Has_A_Marker
    ( Data : IN Change_State;
      Item : IN Object_Value ) RETURN Boolean;

      -- Should return true only if the given item is one of
      -- the special marker designator values.


  FUNCTION Change_Token
    ( Data : IN Change_State;
      Back : IN Simple_Count;
      Size : IN Simple_Count ) RETURN Stored_Array;

      -- Should build a Stored_Array marker value out of the supplied
      -- parameters.  Special care must be taken so that the second
      -- element of the Stored_Array can never be the same special
      -- marker that is in the first element of the array.  It doesn't
      -- matter what the third, fourth, etc.. elements are.  This is
      -- because two adjacent identical special marker values are defined
      -- to represent a single special marker value in the original stream.
      -- However, if the Stored_Array size is greater than three, the
      -- following pattern would be ok because the doubled special marker
      -- would occur "inside" the marker value:   ( M, X, M, M, ... )


  FUNCTION Length_Value
    ( Data : IN Change_State;
      List : IN Stored_Array ) RETURN Simple_Count;

  FUNCTION Backup_Value
    ( Data : IN Change_State;
      List : IN Stored_Array ) RETURN Simple_Count;

      -- These two functions should be the inverse of the Change_Token
      -- function.  The instantiated procedure Verify_Translation can
      -- be used to verify that these two functions are the converse of
      -- Change_Token, and that Change_Token will never generate an
      -- illegal marker value (i.e. ( M, M, X, ... ) ).
      --                              ^  ^  -- if same => error


  PACKAGE Utility IS NEW Compression_Primitives
    ( Object_Value => Object_Value,
      Change_State => Change_State,
      Object_Count => Simple_Count,
      Object_Index => Object_Index,
      Object_Array => Object_Array,
      Stored_Index => Stored_Index,
      Stored_Array => Stored_Array,
      Count_Limits => Count_Values,
      Has_A_Marker => Has_A_Marker,
      Change_Token => Change_Token,
      Backup_Value => Backup_Value,
      Length_Value => Length_Value );


  Utility_Parameters : Utility.Compression_Arguments;



  FUNCTION Has_A_Marker
    ( Data : IN Change_State;
      Item : IN Object_Value ) RETURN Boolean IS
  BEGIN
      FOR Index IN 1 .. Data.Marker_Entry.Size LOOP
          IF Memory_Definition_1750a."="( Item, Data.Marker_Entry.List( Index ))
          THEN
              RETURN True;
          END IF;
      END LOOP;
      RETURN False;
  END Has_A_Marker;



  FUNCTION Change_Token
    ( Data : IN Change_State;
      Back : IN Simple_Count;
      Size : IN Simple_Count ) RETURN Stored_Array IS

      -- Make sure that the same special marker value is never stored in
      -- the first two elements of the encrypted size and offset token.

      From : Compression_Parameters.Marker_Index := 1;
      Move : Simple_Count := Back;

      Half : Simple_Count;
      List : Stored_Array;

  BEGIN
      -- Each successive special marker value indicates a different
      -- length offset to locate the matching pattern.  The block size
      -- is Data.Cutoff_Count.

      -- Example Length Bias:
      -- Special Marker Number 1  =>  0 * Data.Cutoff_Count
      -- Special Marker Number 2  =>  1 * Data.Cutoff_Count
      -- Special Marker Number 3  =>  3 * Data.Cutoff_Count
      -- etc...

      WHILE Move > Data.Cutoff_Count LOOP
          From := Compression_Parameters.Marker_Index'SUCC( From );
          Move := Move - Data.Cutoff_Count;
      END LOOP;

      IF From NOT IN 1 .. Data.Marker_Entry.Size THEN
          -- Sanity check:  Invalid value for Back (too big).
          RAISE Compression_Parameters.Illegal_Token;
      END IF;

      Half := Move / Data.Totals_Count;

      -- Set first position to the special marker value.
      List( 1 ) := Data.Marker_Entry.List( From );

      -- Set second position to the size coordinate plus the high order
      -- bits of the offset coordinate.
      List( 2 ) := Object_Value( Size + ( Half * Data.Length_Count ) );

      -- Set third position the remaining low order bits of the offset value.
      List( 3 ) := Object_Value( Move MOD Data.Totals_Count );

      -- Each marker token has a different numeric mapping to avoid
      -- duplicating the same special marker in the first two positions.
      -- List( 2 ) will never be 0 at this point.

      --   1 .. M    =>  0   .. M-1
      -- M+1 .. Max  =>  M+1 .. Max

      IF Memory_Definition_1750a."<="( List( 2 ), List( 1 ) ) THEN
          List( 2 ) := Memory_Definition_1750a."-"( List( 2 ), 1 );
      END IF;

      -- Now, List( 2 ) /= List( 1 ).

      RETURN List;

  END Change_Token;



  FUNCTION Backup_Value
    ( Data : IN Change_State;
      List : IN Stored_Array ) RETURN Simple_Count IS
      Count : Simple_Count := Simple_Count( List( 2 ) );
      Times : Simple_Count := 0;
      Extra : Simple_Count;
  BEGIN
      -- Restore correct count using the mapping described above.
      --   0 .. M-1  =>    1 .. M
      -- M+1 .. Max  =>  M+1 .. Max

      IF Count <= Simple_Count( List( List'FIRST ) ) THEN
          Count := Count + 1;
      END IF;

      Extra := Count / Data.Length_Count;
      Extra := Extra * Data.Totals_Count; -- High order bits of offset.

      -- Extra Before:              XXXX..YYYY..
      --                            0000..XXXX..
      -- Extra After:      XXXX..   0000..0000..

      -- Determine offset bias based on the special marker table.
      FOR Check IN 1 .. Data.Marker_Entry.Size LOOP
          EXIT WHEN Memory_Definition_1750a."="
            ( Data.Marker_Entry.List( Check ), List( List'FIRST ) );
          Times := Times + Data.Cutoff_Count;
          IF Compression_Parameters."="( Check, Data.Marker_Entry.Size ) THEN
              -- Sanity check: Special marker value not found in table.
              RAISE Compression_Parameters.Invalid_Input;
          END IF;
      END LOOP;

      RETURN Times + Simple_Count( List( 3 ) ) + Extra;
  END Backup_Value;



  FUNCTION Length_Value
    ( Data : IN Change_State;
      List : IN Stored_Array ) RETURN Simple_Count IS
      Count : Simple_Count := Simple_Count( List( 2 ) );
      Extra : Simple_Count;
  BEGIN
      -- Restore correct count using the mapping described above.
      --   0 .. M-1  =>    1 .. M
      -- M+1 .. Max  =>  M+1 .. Max

      IF Count <= Simple_Count( List( List'FIRST ) ) THEN
          Count := Count + 1;
      END IF;

      Extra := Count / Data.Length_Count;
      Extra := Extra * Data.Length_Count; -- Mask off low order bits.

      -- Extra Before:              XXXX..YYYY..
      --                            0000..XXXX..
      -- Extra After:               XXXX..0000..

      RETURN Count - Extra; -- Remove high order bits of offset to get size.
  END Length_Value;



  PROCEDURE Initialize_Package
    ( Compress_Options : IN Compression_Parameters.Compression_Options :=
                            Compression_Parameters.Default_Options ) IS

      -- Compress_Options.Bit_Offset will make sense from 0 .. 7.

      -- A value to represent the number of bits to "steal" from
      -- the size coordinate and add to the offset coordinate.
      Steal_Over   : Simple_Count :=
                     Simple_Count( 2 ** Compress_Options.Bit_Offset );

      -- Maximum value stored in a single storage unit.
      Higher_Count : Simple_Count := Simple_Count( Object_Value'LAST );

      Marker_State : Change_State;

  BEGIN
      -- Save the marker values for use in computing offset biases.
      Marker_State.Marker_Entry := Compress_Options.Markers;

      -- Save a multiplier for calculating an offset using the high order
      -- bits from the second value in the marker token.
      Marker_State.Totals_Count := Higher_Count + 1;

      -- Calculate a mask based on the bit offset for extracting the size
      -- from the low order bits of the second value in the marker token.
      Marker_State.Length_Count := Higher_Count / Steal_Over + 1;

      -- This value represents the offset range (maximum) value for a single
      -- special marker value.  This number is multiplied by a table index
      -- to compute base offsets when multiple marker values are used.
      Marker_State.Cutoff_Count := Higher_Count * Steal_Over;

      Utility.Initialize_Arguments
        ( Config_State => Utility_Parameters,
          Marker_State => Marker_State,

          -- Save the maximum offset count that can be stored in a token.
          Backup_Count => Marker_State.Cutoff_Count *
                          Simple_Count( Compress_Options.Markers.Size ),

          -- Save the maximum length count that can be stored in a token.
          Length_Count => Marker_State.Length_Count - 1,

          Save_Options => Compress_Options );
  END Initialize_Package;



  FUNCTION Get_Count_Limits RETURN Count_Limits IS
      -- Do a type conversion on the array types...
      List : Count_Values := Utility.Get_Count_Limits( Utility_Parameters );
      Find : Count_Limits;
  BEGIN
      FOR Index IN List'RANGE LOOP
          Find( Index ) := Object_Count( List( Index ) );
      END LOOP;
      RETURN Find;
  END Get_Count_Limits;



  PROCEDURE Compress_Data
    ( Source_List : IN Object_Array;
      Source_Head : IN Object_Index;
      Target_List : IN OUT Object_Array;
      Target_Tail : IN OUT Object_Count ) IS
      Target_Tail_Work : Simple_Count := Simple_Count( Target_Tail );
  BEGIN
      Utility.Compress
        ( Config_Data => Utility_Parameters,
          Source_List => Source_List,
          Target_List => Target_List,
          Source_Head => Source_Head,
          Target_Tail => Target_Tail_Work );
       Target_Tail := Object_Count( Target_Tail_Work );
  END Compress_Data;



  PROCEDURE Uncompress_Data
    ( Source_List : IN Object_Array;
      Source_Head : IN Object_Index;
      Target_List : IN OUT Object_Array;
      Target_Tail : IN OUT Object_Count ) IS
      Target_Tail_Work : Simple_Count := Simple_Count( Target_Tail );
  BEGIN
      Utility.Uncompress
        ( Config_Data => Utility_Parameters,
          Source_List => Source_List,
          Target_List => Target_List,
          Source_Head => Source_Head,
          Target_Tail => Target_Tail_Work );
       Target_Tail := Object_Count( Target_Tail_Work );
  END Uncompress_Data;



  PROCEDURE Verify_Translation
    ( Is_Verified : OUT Boolean;
      Last_Backup : OUT Object_Count;
      Last_Length : OUT Object_Count ) IS
      Last_Backup_Work : Simple_Count;
      Last_Length_Work : Simple_Count;
  BEGIN
      Utility.Verify_Translation
        ( Config_Data => Utility_Parameters,
          Is_Verified => Is_Verified,
          Last_Backup => Last_Backup_Work,
          Last_Length => Last_Length_Work );
       Last_Length := Object_Count( Last_Length_Work );
       Last_Backup := Object_Count( Last_Backup_Work );
  END Verify_Translation;



END Compression_Instance;
