
WITH Memory_Definition_1750a;

PACKAGE Compression_Parameters IS

  TYPE Operating_Option IS
    ( Normal_Processing_Mode,
      Determine_Theoretical_Benefit );

  TYPE Length_Parameter IS
    ( Minimum_Backup, Maximum_Backup, Storage_Backup,
      Minimum_Length, Maximum_Length, Storage_Length );


  Input_Comment_Prefix : CONSTANT String := "--";
  Fixed_Code_Separator : CONSTANT String := ">>";


  Data_Overflow : EXCEPTION; -- Raised if target buffer space is exceeded.
  Invalid_Input : EXCEPTION; -- Raised if the compression algorithm fails.
  Illegal_Token : EXCEPTION; -- Raised if an attempt is made to create an
                             -- illegal special token marker.

  SUBTYPE String_Argument IS String( 1 .. 20 );

  TYPE Expansion_Range IS RANGE - 2 ** 15 .. 2 ** 15 - 1;


  TYPE Marker_Index IS RANGE 1 .. 32;
  TYPE Marker_Array IS ARRAY ( Marker_Index RANGE <> ) OF
      Memory_Definition_1750a.Byte_Value;

  TYPE Marker_Items( Size : Marker_Index := 1 ) IS RECORD
      List : Marker_Array( 1 .. Size );
  END RECORD;

  TYPE Compression_Options IS RECORD

      Operate : Operating_Option := Normal_Processing_Mode;
      Markers : Marker_Items := ( Size => 1, List => ( 1 => 0 ) );

      Bit_Offset : Natural := 0;
        -- To encode a repeating pattern, two numbers are needed:
        --   (1) Size of the pattern
        --   (2) Offset position of the pattern (location)
        -- Both of these numbers are encoded into the token marker.
        -- 
        -- The Bit_Offset parameter indicates the number of bits
        -- to take away from the default representation size for
        -- the size coordinate and add to the default representation
        -- size of the offset coordinate.
        -- 
        -- Example:
        -- 
        -- Default Representation: ( same as Bit_Offset == 0 )
        --    (1) Size    =>  8 Bits
        --    (2) Offset  =>  8 Bits

        -- Representation with Bit_Offset == 1
        --    (1) Size    =>  7 Bits
        --    (2) Offset  =>  9 Bits


      -- General purpose arguments...

      -- Currently not used but included here for potential
      -- future use by compression/uncompression algorithms.

      Expansion : Expansion_Range := 0;
      Text_Args : String_Argument := ( OTHERS => ' ' );

  END RECORD;


  Default_Options : CONSTANT Compression_Options :=
    ( Operate => Normal_Processing_Mode,
      Markers => ( Size => 1, List => ( 1 => 0 ) ),
      Bit_Offset => 0,
      Expansion  => 0,
      Text_Args  => ( OTHERS => ' ' ) );


  PROCEDURE Validate_Markers( Options : IN Compression_Options );
  -- Raises Invalid_Input if the marker list contains any illegal duplicates.

  FUNCTION Contains_Comment_Prefix( Text : IN String ) RETURN Boolean;
  FUNCTION Contains_Code_Separator( Text : IN String ) RETURN Boolean;

  FUNCTION Image( Value : Natural ) RETURN String;
  FUNCTION Image( Index : Marker_Index ) RETURN String;

END Compression_Parameters;
