
WITH Text_IO;
WITH Memory_Hash_Table;

PACKAGE BODY Compression_Primitives IS

  -- Note that the values encoded with the passed in function Change_Token
  -- and decoded with the functions Length_Value and Backup_Value only store
  -- values in the range 1 .. Backup_Count and 1 .. Length_Count.

  -- However, these values actually represent the ranges
  --   Minimum_Backup .. Backup_Count + Minimum_Backup - 1
  --   Minimum_Length .. Length_Count + Minimum_Length - 1
  -- since lengths < Minimum_Length and offset < Minimum_Backup are
  -- never used.  These conversions are done in the routines
  -- Set_Actual_Data and Get_Actual_Data.
  --
  -- One special note, if a compressed byte sequence ends with a single
  -- special marker token, this value is simply removed.  This signifies
  -- the special case where an even byte boundry is required (as when
  -- the base unit is actually words), and so the compressed sequence is
  -- an odd size.



  FUNCTION Plus
    ( Index : IN Object_Index;
      Count : IN Object_Count ) RETURN Object_Index IS
  BEGIN
      RETURN Object_Index( Object_Count( Index ) + Count );
  END Plus;



  FUNCTION Minus
    ( Index : IN Object_Index;
      Count : IN Object_Count ) RETURN Object_Index IS
  BEGIN
      RETURN Object_Index( Object_Count( Index ) - Count );
  END Minus;



  FUNCTION Plus
    ( Index : IN Object_Index;
      Count : IN Object_Count;
      Last  : IN Object_Index ) RETURN Object_Index IS
      Found : Object_Index;
  BEGIN
      Found := Object_Index( Object_Count( Index ) + Count );
      IF Found > Last THEN
          Found := Last;
      END IF;
      RETURN Found;
  EXCEPTION
      WHEN Standard.Constraint_Error | Standard.Numeric_Error =>
          RETURN Last;
  END Plus;



  FUNCTION Minus
    ( Index : IN Object_Index;
      Count : IN Object_Count;
      First : IN Object_Index ) RETURN Object_Index IS
      Found : Object_Index;
  BEGIN
      Found := Object_Index( Object_Count( Index ) - Count );
      IF Found < First THEN
          Found := First;
      END IF;
      RETURN Found;
  EXCEPTION
      WHEN Standard.Constraint_Error | Standard.Numeric_Error =>
          RETURN First;
  END Minus;



  FUNCTION Min( A : Object_Count; B : Object_Count ) RETURN Object_Count IS
  BEGIN
      IF A < B THEN
          RETURN A;
      ELSE
          RETURN B;
      END IF;
  END Min;



  PROCEDURE Get_Actual_Data
    ( Config_Data : IN Compression_Arguments;
      Marker_Item : IN Stored_Array;
      Backup_Data : OUT Object_Count;
      Length_Data : OUT Object_Count ) IS
  BEGIN
      -- Convert to Actual Parameters...

      Length_Data :=
          Length_Value( Config_Data.Marker_State, Marker_Item ) +
          Config_Data.Marker_Array( Compression_Parameters.Minimum_Length ) - 1;

      Backup_Data :=
          Backup_Value( Config_Data.Marker_State, Marker_Item ) +
          Config_Data.Marker_Array( Compression_Parameters.Minimum_Backup ) - 1;

  END Get_Actual_Data;



  PROCEDURE Set_Actual_Data
    ( Config_Data : IN Compression_Arguments;
      Marker_Item : OUT Stored_Array;
      Backup_Data : IN Object_Count;
      Length_Data : IN Object_Count ) IS

      Backup_Used : Object_Count := 0;
      Length_Used : Object_Count := 0;
      Marker_List : Stored_Array;

  BEGIN
      Backup_Used := Backup_Data -
                     Config_Data.Marker_Array
                       ( Compression_Parameters.Minimum_Backup ) + 1;

      Length_Used := Length_Data -
                     Config_Data.Marker_Array
                       ( Compression_Parameters.Minimum_Length ) + 1;

      Marker_List := Change_Token
        ( Data => Config_Data.Marker_State,
          Back => Backup_Used,
          Size => Length_Used );

      IF NOT Has_A_Marker
        ( Config_Data.Marker_State, Marker_List( Marker_List'FIRST ) )
      THEN
          RAISE Illegal_Token;
      END IF;

      Marker_Item := Marker_List;

  END Set_Actual_Data;



  PROCEDURE Set_Position
    ( Config_Data    : IN Compression_Arguments;
      Marker_Item    : OUT Stored_Array;
      Source_Minimum : IN Object_Index;
      Source_Maximum : IN Object_Index;
      Copy_From_Head : IN Object_Index;
      Copy_From_Tail : IN Object_Index ) IS

      Maximum_Backup : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Maximum_Backup );
      Minimum_Backup : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Minimum_Backup );

      Actual_Backup : Object_Count;
      Actual_Length : Object_Count;

  BEGIN
      Actual_Length := Object_Count( Copy_From_Tail - Copy_From_Head + 1 );
      Actual_Backup := Object_Count( Source_Maximum - Copy_From_Head + 1 );

      IF Actual_Backup NOT IN Minimum_Backup .. Maximum_Backup THEN
          RAISE Invalid_Input;
      END IF;

      Set_Actual_Data
        ( Config_Data => Config_Data,
          Marker_Item => Marker_Item,
          Backup_Data => Actual_Backup,
          Length_Data => Actual_Length );

  END Set_Position;



  PROCEDURE Get_Position
    ( Config_Data    : IN Compression_Arguments;
      Marker_Data    : IN Stored_Array;
      Source_Minimum : IN Object_Index;
      Source_Maximum : IN Object_Index;
      Copy_From_Head : OUT Object_Index;
      Copy_From_Tail : OUT Object_Index ) IS

      Stored_Length : Object_Count;
      Stored_Backup : Object_Count;

      Use_From_Head : Object_Index;
      Use_From_Tail : Object_Index;

  BEGIN
      -- Convert to Actual Parameters...
      Get_Actual_Data
        ( Config_Data => Config_Data,
          Marker_Item => Marker_Data,
          Backup_Data => Stored_Backup,
          Length_Data => Stored_Length );

      Use_From_Head :=
          Minus( Source_Maximum, Stored_Backup - 1, Source_Minimum );
      Use_From_Tail := Plus( Use_From_Head, Stored_Length ) - 1;

      IF Use_From_Head NOT IN Source_Minimum .. Source_Maximum THEN
          RAISE Invalid_Input;
      END IF;
      IF Use_From_Tail NOT IN Source_Minimum .. Source_Maximum THEN
          RAISE Invalid_Input;
      END IF;

      Copy_From_Head := Use_From_Head;
      Copy_From_Tail := Use_From_Tail;

  END Get_Position;



  PROCEDURE Initialize_Arguments
    ( Config_State : IN OUT Compression_Arguments;
      Marker_State : IN Change_State;
      Backup_Count : IN Object_Count;
      Length_Count : IN Object_Count;
      Save_Options : IN Compression_Parameters.Compression_Options :=
                        Compression_Parameters.Default_Options ) IS

      Use_Backup_Count : Object_Count := Backup_Count;
      Use_Length_Count : Object_Count := Length_Count;

  BEGIN
      Config_State.Initialized  := True;

      -- Check for illegal duplicate marker values:
      Compression_Parameters.Validate_Markers( Save_Options );

      Config_State.Save_Options := Save_Options;
      Config_State.Stored_Count :=
          Object_Count( Stored_Index'LAST - Stored_Index'FIRST + 1 );

      Config_State.Marker_Array :=
        ( Compression_Parameters.Storage_Backup => Use_Backup_Count,
          Compression_Parameters.Storage_Length => Use_Length_Count,
          Compression_Parameters.Minimum_Length =>
              Config_State.Stored_Count + 1,
          Compression_Parameters.Maximum_Length =>
              Use_Length_Count + Config_State.Stored_Count,
          Compression_Parameters.Minimum_Backup =>
              Config_State.Stored_Count + 1,
          Compression_Parameters.Maximum_Backup =>
              Use_Backup_Count + Config_State.Stored_Count );

      Config_State.Marker_State := Marker_State;
  END Initialize_Arguments;
 


  FUNCTION Get_Count_Limits
    ( Config_Data : IN Compression_Arguments )
      RETURN Count_Limits IS
  BEGIN
      RETURN Config_Data.Marker_Array;
  END Get_Count_Limits;
 
 

  PROCEDURE Compress
    ( Config_Data : IN Compression_Arguments;
      Source_List : IN Object_Array;
      Source_Head : IN Object_Index;
      Target_List : IN OUT Object_Array;
      Target_Tail : IN OUT Object_Count ) IS


      Maximum_Backup : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Maximum_Backup );

      Minimum_Backup : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Minimum_Backup );

      Maximum_Length : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Maximum_Length );

      Minimum_Length : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Minimum_Length );


      Target_Next : Object_Index := Target_List'FIRST;
      Source_Next : Object_Index := Source_Head;



      PACKAGE Hash IS NEW Memory_Hash_Table
        ( Memory_Value => Object_Value,
          Memory_Index => Object_Index,
          Memory_Count => Object_Count,
          Memory_Array => Object_Array );



      PROCEDURE Add_Simple_Token( Item : IN Object_Value ) IS
      BEGIN
          IF Target_Next NOT IN Target_List'RANGE THEN
              RAISE Data_Overflow;
          END IF;
          Target_List( Target_Next ) := Item;
          Target_Next := Target_Next + 1;
      END Add_Simple_Token;



      PROCEDURE Add_Literal_Token( Item : IN Object_Value ) IS
      BEGIN
          Add_Simple_Token( Item );
          IF Has_A_Marker( Config_Data.Marker_State, Item ) THEN -- Double it...
              Add_Simple_Token( Item );
          END IF;
      END Add_Literal_Token;



      PROCEDURE Add_Pattern_Token
        ( Find : IN Object_Index;
          Next : IN Object_Index;
          Size : IN Object_Count ) IS
          Marker_Item : Stored_Array;
      BEGIN
          Set_Position
            ( Config_Data    => Config_Data,
              Marker_Item    => Marker_Item,
              Source_Minimum => Source_List'FIRST,
              Source_Maximum => Next - 1,
              Copy_From_Head => Find,
              Copy_From_Tail => Plus( Find, Size ) - 1 );
          FOR Index IN Marker_Item'RANGE LOOP
              Add_Simple_Token( Marker_Item( Index ) );
          END LOOP;
      END Add_Pattern_Token;



      PROCEDURE Find_Longest_Match
        ( Next : IN Object_Index;
          Find : IN OUT Object_Index;
          Size : IN OUT Object_Count ) IS

          Size_Most : CONSTANT Object_Count :=
                      Min( Object_Count( Next - Source_List'FIRST ),
                           Min( Object_Count( 1 + Source_List'LAST - Next ),
                                Maximum_Length ) );

          Test_Most : Object_Count;
          Next_Used : Object_Index;
          Test_Used : Object_Index;

          Head_Minimum : Object_Index;
          Head_Maximum : Object_Index;



          PROCEDURE Act_On( Index : IN Object_Index; Quit : IN OUT Boolean ) IS
              Find_Size : Object_Count := 1;
          BEGIN
              Next_Used := Next + 1;
              Test_Used := Index + 1;

              Test_Most := Min( Size_Most,
                                Object_Count( Next_Used - Test_Used ) );

              FOR Loop_Index IN 2 .. Test_Most LOOP
                  EXIT WHEN Source_List( Next_Used ) /=
                            Source_List( Test_Used );
                  Find_Size := Find_Size + 1;
                  Next_Used := Next_Used + 1;
                  Test_Used := Test_Used + 1;
              END LOOP;

              IF Find_Size > Size THEN
                  Size := Find_Size;
                  Find := Index;
                  Quit := Find_Size = Size_Most;
              END IF;

          END Act_On;


          PROCEDURE Iterate_List IS NEW Hash.Reverse_Iterate_Range( Act_On );


      BEGIN

          Size := 0;
          Find := Next;

          Head_Minimum :=
              Minus( Index => Next,
                     Count => Maximum_Backup,
                     First => Source_List'FIRST );
          Head_Maximum :=
              Minus( Index => Next,
                     Count => Minimum_Backup,
                     First => Source_List'FIRST );

          Iterate_List
            ( Source_List( Next ),
              Minimum => Head_Minimum,
              Maximum => Head_Maximum );
      END Find_Longest_Match;



      PROCEDURE Compress_The_Data IS
          Find : Object_Index := Object_Index'FIRST;
          Size : Object_Count := 0;
      BEGIN
          -- Copy over initial data NOT to be compressed.
          Target_Next := Target_Next + ( Source_Head - Source_List'FIRST );
          Target_List( Target_List'FIRST .. Target_Next - 1 ) :=
              Source_List( Source_List'FIRST .. Source_Head - 1 );

          -- Initialize first few data elements.
          FOR Index IN 1 .. Minimum_Length LOOP

              -- In case the list is very small...
              EXIT WHEN Source_Next NOT IN Source_List'RANGE;

              -- If already have at least the minimum loaded then exit.
              -- This could happen if some values at the head of the list
              -- are not to be compressed, but are available for searching.
              EXIT WHEN Object_Count( Source_Next - Source_List'FIRST + 1 ) >
                        Minimum_Length;

              Add_Literal_Token( Source_List( Source_Next ) );
              Source_Next := Source_Next + 1;

          END LOOP;
    
          Search_Loop:
          WHILE Source_Next IN Source_List'RANGE LOOP

              Find_Longest_Match
                ( Next => Source_Next,
                  Find => Find,
                  Size => Size );
    
              IF Size >= Minimum_Length THEN

                  -- NOTE:  Here, Source_Next - Find is also >= Minimum_Length
                  -- Therefore ( Source_Next - Find ) - Minimum_Length + 1 > 0.

                  Add_Pattern_Token( Find, Source_Next, Size );
                  Source_Next := Plus( Source_Next, Size );

              ELSE
                  Add_Literal_Token( Source_List( Source_Next ) );
                  Source_Next := Source_Next + 1;
              END IF;
    
          END LOOP Search_Loop;
    
          Target_Tail := Object_Count( Target_Next - 1 );

      END Compress_The_Data;



  BEGIN

      IF NOT Config_Data.Initialized THEN
          RAISE Improper_Calling_Sequence;
      END IF;

      FOR Index IN Source_List'RANGE LOOP
          Hash.Store( Index, Source_List( Index ) );
      END LOOP;

      Hash.Done_Adding_Data;

      Compress_The_Data;

  END Compress;



  PROCEDURE Uncompress
    ( Config_Data : IN Compression_Arguments;
      Source_List : IN Object_Array;
      Source_Head : IN Object_Index;
      Target_List : IN OUT Object_Array;
      Target_Tail : IN OUT Object_Count ) IS

      Source_Next : Object_Index := Source_Head;
      Target_Next : Object_Index := Target_List'FIRST;



      PROCEDURE Add_Simple_Token( Item : IN Object_Value ) IS
      BEGIN
          IF Target_Next NOT IN Target_List'RANGE THEN
              RAISE Data_Overflow;
          END IF;
          Target_List( Target_Next ) := Item;
          Target_Next := Target_Next + 1;
      END Add_Simple_Token;



      PROCEDURE Expand_Pattern( Marker : IN Stored_Array ) IS
          Target_Head : Object_Index;
          Target_Tail : Object_Index;
          Target_Last : Object_Index;
      BEGIN
          Get_Position
            ( Config_Data    => Config_Data,
              Marker_Data    => Marker,
              Source_Minimum => Target_List'FIRST,
              Source_Maximum => Target_Next - 1,
              Copy_From_Head => Target_Head,
              Copy_From_Tail => Target_Tail );
          Target_Last := Target_Next + Target_Tail - Target_Head;
          IF Target_Last NOT IN Target_List'RANGE THEN
              RAISE Data_Overflow;
          END IF;
          Target_List( Target_Next .. Target_Last ) :=
              Target_List( Target_Head .. Target_Tail );
          Target_Next := Target_Last + 1;
      END Expand_Pattern;



  BEGIN

      IF NOT Config_Data.Initialized THEN
          RAISE Improper_Calling_Sequence;
      END IF;

      -- Copy directly any header information or any uncompressed data.

      IF Source_Head > Source_List'FIRST THEN
          Target_Next := Target_List'FIRST + (Source_Head - Source_List'FIRST);
          Target_List( Target_List'FIRST .. Target_Next - 1 ) :=
              Source_List( Source_List'FIRST .. Source_Head - 1 );
      END IF;

      -- Begin Decompression

      WHILE Source_Next IN Source_List'RANGE LOOP
          IF Has_A_Marker( Config_Data.Marker_State,
                           Source_List( Source_Next ) )
          THEN

              IF Source_Next = Source_List'LAST THEN
                  -- Ignore n-boundry filler.
                  Source_Next := Source_Next + 1;

              ELSIF Has_A_Marker
                ( Config_Data.Marker_State,
                  Source_List( Source_Next + 1 ) )
              AND THEN
                 Source_List( Source_Next ) =
                 Source_List( Source_Next + 1 )
              THEN
                 -- Is an encoded special marker only if Has_A_Marker is
                 -- true and the next item is exactly the same.  If it
                 -- is a special marker doubled, then replace it with
                 -- a single special marker.

                  Add_Simple_Token( Source_List( Source_Next ) );
                  Source_Next := Source_Next + 2;

              ELSE
                  -- Found a pattern token and so must expand it.
               
                  Trap_Problem: DECLARE
                      Source_Tail : CONSTANT Object_Index :=
                          Plus( Source_Next, Config_Data.Stored_Count ) - 1;
                  BEGIN
                      IF Source_Tail NOT IN Source_List'RANGE THEN
                          -- If not enough room at end of list to contain a
                          -- marker then don't bother, just add literal item.
                          Add_Simple_Token( Source_List( Source_Next ) );
                          Source_Next := Source_Next + 1;

                      ELSE
                          Expand_Pattern
                            ( Stored_Array
                                ( Source_List( Source_Next .. Source_Tail ) ) );
                          Source_Next := Source_Tail + 1;
                      END IF;
                  EXCEPTION
                      WHEN Standard.Constraint_Error | Standard.Numeric_Error =>
                          RAISE Invalid_Input;
                  END Trap_Problem;
              END IF;
          ELSE
              Add_Simple_Token( Source_List( Source_Next ) );
              Source_Next := Source_Next + 1;
          END IF;
      END LOOP;
      Target_Tail := Object_Count( Target_Next - 1 );
  END Uncompress;



  PROCEDURE Verify_Translation
    ( Config_Data : IN Compression_Arguments;
      Is_Verified : OUT Boolean;
      Last_Backup : OUT Object_Count;
      Last_Length : OUT Object_Count ) IS

      TYPE Problem_Type IS
        ( None, Invalid_Token, Invalid_Match,
          Invalid_Scheme, Invalid_Back, Invalid_Size );

      Abort_Processing : EXCEPTION;


      Minimum_Backup : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Minimum_Backup );

      Maximum_Backup : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Maximum_Backup );

      Minimum_Length : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Minimum_Length );

      Maximum_Length : CONSTANT Object_Count :=
          Config_Data.Marker_Array
            ( Compression_Parameters.Maximum_Length );



      PROCEDURE Print( Text : String ) IS
      BEGIN
          Text_IO.Put_Line( Text );
      END Print;



      PROCEDURE Check
        ( Set_Back : IN Object_Count;
          Set_Size : IN Object_Count ) IS
          Mode : Problem_Type := None;
          Marker_Item : Stored_Array;
          Get_Back : Object_Count;
          Get_Size : Object_Count;
      BEGIN

          Set_Actual_Data
            ( Config_Data => Config_Data,
              Marker_Item => Marker_Item,
              Backup_Data => Set_Back,
              Length_Data => Set_Size );

          Get_Actual_Data
            ( Config_Data => Config_Data,
              Marker_Item => Marker_Item,
              Backup_Data => Get_Back,
              Length_Data => Get_Size );

          IF Marker_Item'LENGTH < 2 THEN
              Mode := Invalid_Scheme;
          ELSIF Set_Back /= Get_Back THEN
              Mode := Invalid_Back;
          ELSIF Set_Size /= Get_Size THEN
              Mode := Invalid_Size;
          ELSIF NOT Has_A_Marker
            ( Config_Data.Marker_State, Marker_Item( Marker_Item'FIRST ) )
          THEN
              Mode := Invalid_Token;
          ELSIF Marker_Item( Marker_Item'FIRST ) =
                Marker_Item( Marker_Item'FIRST + 1 )
          THEN
              Mode := Invalid_Match;
          END IF;
          IF Mode /= None THEN

              Is_Verified := False;
              Last_Backup := Set_Back;
              Last_Length := Set_Size;

              Print( "Verification Failed => " & Problem_Type'IMAGE( Mode ) );
              Print( "" );
              Print( "  Set Back " & Object_Count'IMAGE( Set_Back ) );
              Print( "  Set Size " & Object_Count'IMAGE( Set_Size ) );
              Print( "" );
              FOR Index IN Marker_Item'RANGE LOOP
                  Print( "Marker(" & Stored_Index'IMAGE( Index ) & " ) => " &
                         Object_Value'IMAGE( Marker_Item( Index ) ) );
              END LOOP;
              Print( "" );
              Print( "  Get Back " & Object_Count'IMAGE( Get_Back ) );
              Print( "  Get Size " & Object_Count'IMAGE( Get_Size ) );
              Print( "" );

              RAISE Abort_Processing;

          END IF;
      END Check;



      PROCEDURE Perform_Tests
        ( Min_Back : Object_Count;
          Max_Back : Object_Count;
          Min_Size : Object_Count;
          Max_Size : Object_Count ) IS
          Max_Done : Object_Count;
      BEGIN
          FOR Back IN Min_Back .. Max_Back LOOP
              IF Back > Max_Size THEN
                  Max_Done := Max_Size;
              ELSE
                  Max_Done := Back;
              END IF;
              FOR Size IN Min_Size .. Max_Done LOOP
                  Check( Set_Back => Back, Set_Size => Size );
              END LOOP;
          END LOOP;
      END Perform_Tests;



  BEGIN

      IF NOT Config_Data.Initialized THEN
          RAISE Improper_Calling_Sequence;
      END IF;

      Is_Verified := True;
      Last_Backup := Maximum_Backup;
      Last_Length := Maximum_Length;

      Perform_Tests
        ( Min_Back => Minimum_Backup,
          Max_Back => Maximum_Backup,
          Min_Size => Minimum_Length,
          Max_Size => Maximum_Length );

  EXCEPTION
      WHEN Abort_Processing =>
          NULL;
  END Verify_Translation;



END Compression_Primitives;
