
WITH Compression_Parameters;

GENERIC

  TYPE Object_Value IS RANGE <>;
  TYPE Change_State IS PRIVATE;

  TYPE Object_Count IS RANGE <>; -- Must include 0 (Zero).
  TYPE Object_Index IS RANGE <>; -- Any subtype of Object_Count.

  TYPE Object_Array IS ARRAY( Object_Index RANGE <> ) OF Object_Value;
    -- The array of the kind of data to be compressed/uncompressed.

  TYPE Count_Limits IS ARRAY( Compression_Parameters.Length_Parameter )
                    OF Object_Count;
    -- Compression parameters.

  TYPE Stored_Index IS RANGE <>;
  TYPE Stored_Array IS ARRAY( Stored_Index ) OF Object_Value;
    -- Number of items needed to store counts of token sizes and
    -- array location offset counts.  Must be at least of length 2.

  WITH FUNCTION Has_A_Marker
    ( Data : IN Change_State;
      Item : IN Object_Value ) RETURN Boolean;
    -- Return true if the item is, or contains, special marker bits.

  WITH FUNCTION Change_Token
    ( Data : IN Change_State;
      Back : IN Object_Count;
      Size : IN Object_Count ) RETURN Stored_Array IS <>;
      -- Must ALWAYS return a value that has a marker in the first element
      -- of the array result.  Must NEVER return a value that has a marker
      -- in the second element of the array result.

  WITH FUNCTION Backup_Value
    ( Data : IN Change_State;
      List : IN Stored_Array ) RETURN Object_Count IS <>;
      -- 1 of 2 inverse functions for Change_Token.

  WITH FUNCTION Length_Value
    ( Data : IN Change_State;
      List : IN Stored_Array ) RETURN Object_Count IS <>;
      -- 2 of 2 inverse functions for Change_Token.

PACKAGE Compression_Primitives IS

  Data_Overflow : EXCEPTION RENAMES Compression_Parameters.Data_Overflow;
                  -- Raised if Target buffer space is exceeded.
  Invalid_Input : EXCEPTION RENAMES Compression_Parameters.Invalid_Input;
                  -- Raised if the compression algorithm fails.
  Illegal_Token : EXCEPTION RENAMES Compression_Parameters.Illegal_Token;
                  -- Raised in Compress_Data if an attempt is
                  -- made to create a special token with a marker
                  -- in the first array element.

  Improper_Calling_Sequence : EXCEPTION;


  TYPE Compression_Arguments IS PRIVATE;



  PROCEDURE Initialize_Arguments
    ( Config_State : IN OUT Compression_Arguments;
      Marker_State : IN Change_State;
      Backup_Count : IN Object_Count;   -- Maximum reverse offset search range.
      Length_Count : IN Object_Count;   -- Maximum pattern matching length.
      Save_Options : IN Compression_Parameters.Compression_Options :=
                        Compression_Parameters.Default_Options );


  FUNCTION Get_Count_Limits
    ( Config_Data : IN Compression_Arguments )
      RETURN Count_Limits;



  PROCEDURE Compress
    ( Config_Data : IN Compression_Arguments;
      Source_List : IN Object_Array;
      Source_Head : IN Object_Index;
      Target_List : IN OUT Object_Array;
      Target_Tail : IN OUT Object_Count );


  PROCEDURE Uncompress
    ( Config_Data : IN Compression_Arguments;
      Source_List : IN Object_Array;
      Source_Head : IN Object_Index;
      Target_List : IN OUT Object_Array;
      Target_Tail : IN OUT Object_Count );


  PROCEDURE Verify_Translation
    ( Config_Data : IN Compression_Arguments;
      Is_Verified : OUT Boolean;
      Last_Backup : OUT Object_Count;   -- First instance of a violation.
      Last_Length : OUT Object_Count ); -- First instance of a violation.


PRIVATE

  TYPE Compression_Arguments IS RECORD
      Save_Options : Compression_Parameters.Compression_Options :=
                     Compression_Parameters.Default_Options;
      Stored_Count : Object_Count := 2;
      Marker_Array : Count_Limits := ( OTHERS => 0 );
      Marker_State : Change_State;
      Initialized  : Boolean := False;
  END RECORD;

END Compression_Primitives;
