
WITH Text_IO;
WITH Calendar;
WITH Compression_Statistics;
WITH Compression_Parameters;
WITH Memory_Definition_1750A;
WITH Memory_Supporting_1750A;

PACKAGE BODY Compression_Support IS

  -- SPECIAL NOTE:  Managing Word Boundries
  --
  --   This particular instance of the compression algorithm operates
  --   on byte streams.  However, the format of the input memory map
  --   file is in words.  The compression algorithm reads a memory map
  --   file, compresses it, and creates another memory map file.  If
  --   the compressed memory map file ends on an odd byte boundry, then
  --   a special marker token will be used to create a full word in the
  --   last position.  Then, the decompression algorithm will ignore the
  --   last byte of an input byte stream if it is a special marker value.

  SUBTYPE Number_Of_Bits IS Natural RANGE 0 .. 7;

  Maximum_Hundreths : CONSTANT := 24 * 60 * 60 * 100;
  Minimum_Hundreths : CONSTANT := - Maximum_Hundreths;

  TYPE Hundreths_Range IS RANGE Minimum_Hundreths .. Maximum_Hundreths;

  Debug : CONSTANT Boolean := False;

  Maximum_Columns : CONSTANT := 8;

  SUBTYPE File_Name_Length IS Natural RANGE 0 .. 132;

  TYPE File_Name_Record( Size : File_Name_Length := 0 ) IS RECORD
      Text : String( 1 .. Size );
  END RECORD;

  TYPE File_Name_Data IS RECORD
      File : Text_IO.File_Type;
      Name : File_Name_Record;
  END RECORD;

  TYPE File_Name_Kind IS
    ( Compression_Attributes,
      Compression_Input_File,
      Compression_Output_File );

  File_State : ARRAY( File_Name_Kind ) OF File_Name_Data;

  Parameters : Compression_Parameters.Compression_Options :=
               Compression_Parameters.Default_Options;

  Start_Time : Calendar.Time := Calendar.Clock;
  Stop_Time  : Calendar.Time := Start_Time;



  FUNCTION "+"( X, Y : Compression_Instance.Object_Index )
      RETURN Compression_Instance.Object_Index
      RENAMES Memory_Definition_1750a."+";

  FUNCTION "-"( X, Y : Compression_Instance.Object_Index )
      RETURN Compression_Instance.Object_Index
      RENAMES Memory_Definition_1750a."-";

  FUNCTION "/"( X, Y : Compression_Instance.Object_Index )
      RETURN Compression_Instance.Object_Index
      RENAMES Memory_Definition_1750a."/";

  FUNCTION "*"( X, Y : Compression_Instance.Object_Index )
      RETURN Compression_Instance.Object_Index
      RENAMES Memory_Definition_1750a."*";

  FUNCTION ">="( X, Y : Compression_Instance.Object_Index )
      RETURN Boolean RENAMES Memory_Definition_1750a.">=";

  FUNCTION "<"( X, Y : Compression_Instance.Object_Count )
      RETURN Boolean RENAMES Memory_Definition_1750a."<";



  PACKAGE Statistics IS NEW Compression_Statistics
    ( Compression_Instance.Object_Index, Print );



  PROCEDURE Print( Text : IN String ) IS
  BEGIN
      Text_IO.Put_Line( Text );
  END Print;



  PROCEDURE Error( Text : IN String ) IS
  BEGIN
      Print( "***ERROR*** " & Text );
  END Error;



  PROCEDURE Print_Help_Text IS
  BEGIN
      Print( " " );
      Print( "To encode a repeating pattern, two numbers are needed:" );
      Print( "  (1) Size of the pattern" );
      Print( "  (2) Offset position of the pattern (location)" );
      Print( " " );
      Print( "Both of these numbers are encoded into the token marker." );
      Print( "The Offset Bits parameter indicates the number of bits" );
      Print( "to take away from the default representation size for" );
      Print( "the size coordinate and add to the default representation" );
      Print( "size of the offset coordinate." );
      Print( " " );
      Print( "Examples:" );
      Print( " " );
      Print( "Default Representation: ( same as Offset Bits == 0 ) " );
      Print( "   (1) Size    =>  8 Bits" );
      Print( "   (2) Offset  =>  8 Bits" );
      Print( " " );
      Print( "Representation with Offset Bits == 1" );
      Print( "   (1) Size    =>  7 Bits" );
      Print( "   (2) Offset  =>  9 Bits" );
      Print( " " );
  END Print_Help_Text;



  GENERIC
      WITH PROCEDURE Print( Text : String );
  PROCEDURE Print_Parameter_Format;

  PROCEDURE Print_Parameter_Format IS
  BEGIN
      Print( "" );
      Print( "--------------------------------------" );
      Print( "-- Correct format of parameter file --" );
      Print( "--------------------------------------" );
      Print( "<m>" );
      Print( "<n>" );
      Print( "<b(1)>" );
      Print( "<b(2)>" );
      Print( "<...>" );
      Print( "<b(n)>" );
      Print( "--------------------------------------" );
      Print( "" );
      Print( "Here <m> is an integer denoting the number of bits" );
      Print( "to use in the length versus size representation of" );
      Print( "the marker tokens.  For example, <m> = 0 would use" ); 
      Print( "  <m> = 0:  Size =  8 Bits, Length =  8 Bits" );
      Print( "  <m> = 1:  Size =  7 Bits, Length =  9 Bits" );
      Print( "  <m> = 2:  Size =  6 Bits, Length = 10 Bits" );
      Print( "  <m> = 3:  Size =  5 Bits, Length = 11 Bits" );
      Print( "" );
      Print( "Without changing the implementation of the encoding" );
      Print( "functions, using less than 8 bits for the length is" );
      Print( "not currently supported (ie. negative values for <m>)" );
      Print( "" );
      Print( "And <n> is the number of marker byte values, followed" );
      Print( "by the marker byte values to be used (<b(1)> .. <b(n)>)." );
      Print( "" );
  END Print_Parameter_Format;



  PROCEDURE Close_File( Kind : IN File_Name_Kind ) IS
  BEGIN
      Text_IO.Close( File_State( Kind ).File );
  EXCEPTION
      WHEN OTHERS =>
          NULL; -- ignore any errors...
  END Close_File;



  FUNCTION Query_User( Prompt : IN String ) RETURN String IS
      Text  : String( 1 .. File_Name_Length'LAST );
      Last  : Natural := 0;
  BEGIN
      Text_IO.Put( Prompt );
      Text_IO.Get_Line( Text, Last );
      Text_IO.Put_Line( Text( Text'FIRST .. Last ) );
      RETURN Text( Text'FIRST .. Last );
  END Query_User;



  GENERIC
      TYPE Item_Value IS ( <> );
  FUNCTION Return_Value( Prompt : IN String ) RETURN Item_Value;

  FUNCTION Return_Value( Prompt : IN String ) RETURN Item_Value IS
      Found : CONSTANT String := Query_User( Prompt );
  BEGIN
      IF Found'LENGTH > 0 AND THEN Found( Found'FIRST ) = '?' THEN
          Print_Help_Text;
          RETURN Return_Value( Prompt );
      ELSE
          RETURN Item_Value'VALUE( Found );
      END IF;
  EXCEPTION
      WHEN Standard.Constraint_Error =>
          Print( "" );
          Print_Help_Text;
          Print( "" );
          Error( "Invalid argument from command line." );
          Error( "Expected value in range: " &
                 Item_Value'IMAGE( Item_Value'FIRST ) & " .. " &
                 Item_Value'IMAGE( Item_Value'LAST ) );
          RAISE Unable_To_Initialize;
  END Return_Value;



  GENERIC
      Item_Label : IN String;
      TYPE Item_Value IS ( <> );
  FUNCTION Convert_Value( Text : IN String ) RETURN Item_Value;

  FUNCTION Convert_Value( Text : IN String ) RETURN Item_Value IS
      PROCEDURE Display_Parameter_Format IS NEW Print_Parameter_Format( Error );
  BEGIN
      RETURN Item_Value'VALUE( Text );
  EXCEPTION
      WHEN Standard.Constraint_Error =>
          Error( "Reading " & Item_Label & ": '" & Text & "'" );
          Error( "Expected value in range: " &
                 Item_Value'IMAGE( Item_Value'FIRST ) & " .. " &
                 Item_Value'IMAGE( Item_Value'LAST ) );
          Error( "" );
          Error( "Could not initialize compression algorithms." );
          Error( "" );
          Display_Parameter_Format;
          RAISE Unable_To_Initialize;
  END Convert_Value;



  PROCEDURE Read_Line
    ( From : IN File_Name_Kind;
      Text : IN OUT String;
      Last : IN OUT Natural;
      Head : IN OUT Natural;
      Skip : IN OUT Boolean ) IS
  BEGIN
      Text_IO.Get_Line( File_State( From ).File, Text, Last );
      Head := Text'FIRST;
      WHILE Last IN Text'RANGE AND THEN Text( Last ) = ' ' LOOP
          Last := Last - 1;
      END LOOP;
      WHILE Head <= Last AND THEN Text( Head ) = ' ' LOOP
          Head := Head + 1;
      END LOOP;
      Skip := Compression_Parameters.Contains_Comment_Prefix
               ( Text( Head .. Last ) );
  EXCEPTION
      WHEN Text_IO.Status_Error =>
          Error( "Could not read from file " & File_State( From ).Name.Text );
          Error( "Was Opened => " & Boolean'IMAGE
                   ( Text_IO.Is_Open( File_State( From ).File ) ) );
          Error( "File Mode  => " &
                 Text_IO.File_Mode'IMAGE
                   ( Text_IO.Mode( File_State( From ).File ) ) );
          Error( "File Name  => " &
                 Text_IO.Name( File_State( From ).File ) );
          RAISE Unable_To_Initialize;
  END Read_Line;



  PROCEDURE Next_Word
    ( Line : IN String;
      Line_Head : IN OUT Natural;
      Line_Tail : IN OUT Natural ) IS
  BEGIN
      WHILE Line_Head IN Line'RANGE AND THEN Line( Line_Head ) = ' ' LOOP
          Line_Head := Line_Head + 1;
      END LOOP;
      IF Compression_Parameters.Contains_Comment_Prefix
          ( Line( Line_Head .. Line'LAST ) )
      THEN
          Line_Head := Line'LAST + 1;
      END IF;
      Line_Tail := Line_Head - 1;
      WHILE Line_Tail + 1 IN Line'RANGE AND THEN Line( Line_Tail+1 ) /= ' ' LOOP
          Line_Tail := Line_Tail + 1;
      END LOOP;
  END Next_Word;



  FUNCTION Hex( Index : IN Compression_Instance.Object_Index ) RETURN String IS
  BEGIN
      RETURN Memory_Supporting_1750A.Hex_Image
               ( Item => Memory_Supporting_1750A.Compute_Range( Index ),
                 Size => 5,
                 Pads => ' ' );
  END Hex;



  FUNCTION Hex( Index : IN Compression_Instance.Object_Value ) RETURN String IS
  BEGIN
      RETURN Memory_Supporting_1750A.Hex_Image
               ( Item => Memory_Supporting_1750A.Compute_Range( Index ),
                 Size => 2,
                 Pads => '0' );
  END Hex;



  FUNCTION Image( Index : IN Compression_Instance.Object_Count )
      RETURN String IS
  BEGIN
      RETURN Memory_Supporting_1750A.Image( Index );
  END Image;
  


  FUNCTION Image( Time : Standard.Duration ) RETURN String IS
      Duration_100 : CONSTANT Standard.Duration := 100.00;
      Tick : CONSTANT Hundreths_Range := Hundreths_Range( Time * Duration_100 );
      Text : CONSTANT String := Hundreths_Range'IMAGE( Tick );
      Plus : CONSTANT Natural := 10 - Text'LENGTH;
      Look : String( 1 .. Text'LENGTH + Plus ) :=
             String'( 1 .. Plus => ' ' ) & Text;
  BEGIN
      FOR Index IN Look'LAST - 2 .. Look'LAST LOOP
          IF Look( Index ) = ' ' THEN
              Look( Index ) := '0';
          END IF;
      END LOOP;
      RETURN Look( Look'FIRST .. Look'LAST - 2 ) & "." &
             Look( Look'LAST - 1 .. Look'LAST ) & " Seconds";
  END Image;



  PROCEDURE Dump
    ( List : IN Compression_Instance.Object_Array;
      Head : IN Compression_Instance.Object_Index;
      Mode : IN String ) IS
  BEGIN
      IF Debug THEN
          Print( "" );
          FOR Index IN List'RANGE LOOP
              IF Index IN Head .. Head THEN
                  Print( Compression_Parameters.Fixed_Code_Separator &
                         " Start Here..." );
              END IF;
              Print( Mode & " " & Hex( Index ) & " => " & Hex( List( Index )));
          END LOOP;
          Print( "" );
      END IF;
  END Dump;



  PROCEDURE Initialize_Compression
    ( Label                : IN String;
      Read_Parameters_Only : IN Boolean := False;
      Will_Save_Parameters : IN Boolean := False ) IS
 

      Text : String( 1 .. 132 ) := ( OTHERS => '.' );
      Last : Natural := Text'FIRST - 1;

      Save : Natural := Text'FIRST;

      Save_Head : Natural := Text'FIRST;
      Save_Tail : Natural := Text'FIRST - 1;



      FUNCTION Get_Word RETURN String IS
          Skip : Boolean := False;
      BEGIN
          Save_Head := Save;
          LOOP
              IF Save > Last THEN
                  Read_Line
                    ( From => Compression_Attributes,
                      Text => Text,
                      Last => Last,
                      Head => Save,
                      Skip => Skip );
                  Save_Head := Save;
              END IF;
              Next_Word( Text( Text'FIRST .. Last ), Save_Head, Save_Tail );
              Save := Save_Tail + 1;
              EXIT WHEN Save_Head <= Save_Tail;
          END LOOP;
          RETURN Text( Save_Head .. Save_Tail );
      END Get_Word;



      PROCEDURE Get_Parameters IS

          Size : Compression_Parameters.Marker_Index;
          Bits : Number_Of_Bits;

          FUNCTION Get_Offset IS NEW Return_Value( Number_Of_Bits );
          FUNCTION Get_Marker IS NEW Return_Value
            ( Compression_Parameters.Marker_Index );

      BEGIN
          Bits := Get_Offset( "Using Offset Bits? " );
          Size := Get_Marker( "Number of Markers? " );
          Parameters.Bit_Offset := Bits;
          Parameters.Markers := ( Size => Size, List => ( 1 .. Size => 0 ) );
      END Get_Parameters;



      PROCEDURE Set_Parameters IS

          Size : Compression_Parameters.Marker_Index;

          FUNCTION Get_Offset IS NEW Convert_Value
            ( "Bit Offset", Number_Of_Bits );
          FUNCTION Get_Marker IS NEW Convert_Value
            ( "Marker Count", Compression_Parameters.Marker_Index );
          FUNCTION Get_Byte IS NEW Convert_Value
            ( "Byte Value", Memory_Definition_1750a.Byte_Value );

      BEGIN
          Parameters.Bit_Offset := Get_Offset( Get_Word );
          Size := Get_Marker( Get_Word );
          Parameters.Markers := ( Size => Size, List => ( 1 .. Size => 0 ) );
          FOR Index IN 1 .. Size LOOP
              Parameters.Markers.List( Index ) := Get_Byte( Get_Word );
          END LOOP;
      EXCEPTION
          WHEN Unable_To_Initialize =>
              Error( "Parsing Line: " & Text( Text'FIRST .. Last ) );
              Error( "Parsing Word: " &
                     String'( Text'FIRST .. Save_Head - 1 => ' ' ) &
                     String'( Save_Head .. Save_Tail => '^' ) );
              RAISE;
      END Set_Parameters;



      PROCEDURE Read_Arguments IS

          Found_Error : Boolean := False;


          PROCEDURE Set_Name
            ( Data  : IN OUT File_Name_Data;
              Query : IN String ) IS
              Text  : CONSTANT String := Query_User( Query );
          BEGIN
              Data.Name := ( Size => Text'LENGTH, Text => Text );
          END Set_Name;


          PROCEDURE Open_File
            ( Data  : IN OUT File_Name_Data;
              Query : IN String ) IS
          BEGIN
              Set_Name( Data, Query );
              Text_IO.Open( Data.File, Text_IO.In_File, Data.Name.Text );
          EXCEPTION
              WHEN Text_IO.Name_Error => 
                  Error( "Could not find file: " & Data.Name.Text );
                  Found_Error := True;
              WHEN OTHERS =>
                  Error( "Unknown error opening file: " & Data.Name.Text );
                  Found_Error := True;
          END Open_File;


          PROCEDURE Make_File
            ( Data  : IN OUT File_Name_Data;
              Query : IN String ) IS
          BEGIN
              Set_Name( Data, Query );
              Text_IO.Create( Data.File, Text_IO.Out_File, Data.Name.Text );
          EXCEPTION
              WHEN Text_IO.Name_Error => 
                  Error( "Could not create file: " & Data.Name.Text );
                  Found_Error := True;
              WHEN OTHERS =>
                  Error( "Unknown error creating file: " & Data.Name.Text );
                  Found_Error := True;
          END Make_File;


      BEGIN
          IF Will_Save_Parameters THEN
              Make_File( File_State( Compression_Attributes ),
                         "Compression Parameters? (>) " );
              Open_File( File_State( Compression_Input_File ),
                         "Compression Input File? (<) " );
          ELSE
              Open_File( File_State( Compression_Attributes ),
                         "Compression Parameters? (<) " );
              IF NOT Read_Parameters_Only THEN
                  Open_File( File_State( Compression_Input_File ),
                             "Compression Input File? (<) " );
                  Make_File( File_State( Compression_Output_File ),
                             "Compression Write File? (>) " );
              END IF;
          END IF;
          IF Found_Error THEN
              FOR Kind IN File_Name_Kind LOOP
                  Close_File( Kind );
              END LOOP;
              RAISE Unable_To_Initialize;
          END IF;
      END Read_Arguments;



  BEGIN
      Print( Label & " Version " & Version );
      Read_Arguments;
      IF Will_Save_Parameters THEN
          Get_Parameters;
      ELSE
          Set_Parameters;
          Close_File( Compression_Attributes );
          Compression_Instance.Initialize_Package( Parameters );
      END IF;
  END Initialize_Compression;



  PROCEDURE Save_Parameters
    ( Source_List : IN Compression_Instance.Object_Array;
      Source_Head : IN Compression_Instance.Object_Index;
      Source_Tail : IN Compression_Instance.Object_Index ) IS

      Histogram : ARRAY( Compression_Instance.Object_Value )
                  OF Compression_Instance.Object_Count := ( OTHERS => 0 );


      PROCEDURE Output( Text : IN String ) IS
      BEGIN
          Text_IO.Put_Line( File_State( Compression_Attributes ).File, Text );
      END Output;


      PROCEDURE Dump_Line( Text : IN String ) IS
      BEGIN
          Output( Compression_Parameters.Input_Comment_Prefix & " " & Text );
      END Dump_Line;


      FUNCTION Get_Least_Frequently_Used_Byte
          RETURN Compression_Instance.Object_Value IS

          Found : Compression_Instance.Object_Value :=
                  Compression_Instance.Object_Value'FIRST;
          Count : Compression_Instance.Object_Count :=
                  Compression_Instance.Object_Count'LAST;

      BEGIN
          FOR Index IN Histogram'RANGE LOOP
              IF Histogram( Index ) < Count THEN
                  Count := Histogram( Index );
                  Found := Index;
              END IF;
          END LOOP;

          -- Make sure you don't find it again...
          Histogram( Found ) := Compression_Instance.Object_Count'LAST;

          Print( "Found value " & Memory_Supporting_1750A.Image( Found ) &
                 " with frequency " & Memory_Supporting_1750A.Image( Count ) );

          RETURN Found;

      END Get_Least_Frequently_Used_Byte;


      PROCEDURE Dump_Parameter_Format IS NEW
          Print_Parameter_Format( Dump_Line );


  BEGIN

      Print( "Computing byte frequencies..." );

      FOR Index IN Source_List'FIRST .. Source_Tail LOOP
          Histogram( Source_List( Index ) ) :=
              Compression_Instance.Object_Count'SUCC
                ( Histogram( Source_List( Index ) ) );
      END LOOP;

      FOR Index IN Parameters.Markers.List'RANGE LOOP
          Parameters.Markers.List( Index ) := Get_Least_Frequently_Used_Byte;
      END LOOP;

      Output( Compression_Parameters.Image( Parameters.Bit_Offset ) &
              " " & Compression_Parameters.Input_Comment_Prefix &
              " Add to Offset Bits" );
      Output( Compression_Parameters.Image( Parameters.Markers.Size ) &
              " " & Compression_Parameters.Input_Comment_Prefix &
              " Number of Markers" );
      FOR Index IN Parameters.Markers.List'RANGE LOOP
          Output( Memory_Supporting_1750a.Image
                    ( Parameters.Markers.List( Index ) ) &
                  " " & Compression_Parameters.Input_Comment_Prefix &
                  " Marker Number" &
                  Compression_Parameters.Image( Index ) );
      END LOOP;

      Dump_Parameter_Format;

      Close_File( Compression_Attributes );

  END Save_Parameters;



  PROCEDURE Start_Timer IS
  BEGIN
      Start_Time := Calendar.Clock;
  END Start_Timer;



  PROCEDURE Stop_Timer IS
  BEGIN
      Stop_Time := Calendar.Clock;
  END Stop_Timer;



  PROCEDURE Print_Report
    ( Label       : IN String;
      Source_List : IN Compression_Instance.Object_Array;
      Source_Head : IN Compression_Instance.Object_Index;
      Source_Tail : IN Compression_Instance.Object_Index;
      Target_List : IN Compression_Instance.Object_Array;
      Target_Head : IN Compression_Instance.Object_Index;
      Target_Tail : IN Compression_Instance.Object_Index ) IS



      FUNCTION Elapsed_Time RETURN String IS
      BEGIN
          RETURN Image( Calendar."-"( Stop_Time, Start_Time ) );
      EXCEPTION
          WHEN OTHERS =>
              RETURN "****.**";
      END Elapsed_Time;


      FUNCTION Compression_Ratio
        ( Went_From_Size : IN Compression_Instance.Object_Index;
          Down_To_Length : IN Compression_Instance.Object_Index )
          RETURN String IS
      BEGIN
          IF Went_From_Size >= Down_To_Length THEN
              RETURN "--> " &
                  Statistics.Ratio_Image
                    ( Went_From_Size => Went_From_Size,
                      Down_To_Length => Down_To_Length );
          ELSE
              RETURN "--< " &
                  Statistics.Ratio_Image
                    ( Went_From_Size => Down_To_Length,
                      Down_To_Length => Went_From_Size );
          END IF;
      END Compression_Ratio;


      PROCEDURE Print_Statistics
        ( Mode           : IN String;
          Went_From_Size : IN Compression_Instance.Object_Index;
          Down_To_Length : IN Compression_Instance.Object_Index ) IS
      BEGIN
          IF Went_From_Size >= Down_To_Length THEN
              Print( Mode & ": (Compressing)" );
              Print( "" );
              Statistics.Print_Report
                ( Went_From_Size => Went_From_Size,
                  Down_To_Length => Down_To_Length,
                  Use_Table_Head => "  " );
          ELSE
              Print( Mode & ": (Uncompressing)" );
              Print( "" );
              Statistics.Print_Report
                ( Went_From_Size => Down_To_Length,
                  Down_To_Length => Went_From_Size,
                  Use_Table_Head => "  " );
          END IF;
          Print( "" );
      END Print_Statistics;



  BEGIN
      Print( "" );
      Print( Label );
      Print( "" );
      Print( "  Source First => " & Image( Source_List'FIRST ) );
      Print( "  Source Break => " & Image( Source_Head ) );
      Print( "  Source Last  => " & Image( Source_Tail ) );
      Print( "" );
      Print( "  Target First => " & Image( Target_List'FIRST ) );
      Print( "  Target Break => " & Image( Target_Head ) );
      Print( "  Target Last  => " & Image( Target_Tail ) );
      Print( "" );
      Print( "  Elapsed Time => " & Elapsed_Time );
      Print( "" );
      Print_Statistics
        ( Mode           => "Compressed Code Only",
          Went_From_Size => ( Source_Tail - Source_Head + 1 ),
          Down_To_Length => ( Target_Tail - Target_Head + 1 ) );
      Print_Statistics
        ( Mode           => "Entire Address Space",
          Went_From_Size => ( Source_Tail - Source_List'FIRST + 1 ),
          Down_To_Length => ( Target_Tail - Target_List'FIRST + 1 ) );
      Print( "" );
      Print( Compression_Ratio
              ( Went_From_Size => ( Source_Tail - Source_List'FIRST + 1 ),
                Down_To_Length => ( Target_Tail - Target_List'FIRST + 1 ) ) &
             " " & Elapsed_Time &
             " " & Compression_Parameters.Image( Parameters.Bit_Offset ) &
             " Bits  " &
             Compression_Parameters.Image( Parameters.Markers.Size ) &
             " Markers" );
  END Print_Report;



  PROCEDURE Load_Memory
    ( List : IN OUT Compression_Instance.Object_Array;
      Head : OUT Compression_Instance.Object_Index;
      Last : OUT Compression_Instance.Object_Index ) IS

      Start_At     : Compression_Instance.Object_Index := List'FIRST;
      Get_Head     : Compression_Instance.Object_Index := List'FIRST;
      Set_Head_Now : Boolean := False;



      PROCEDURE Process_Line( Line : IN String ) IS
          Line_Head : Natural := Line'FIRST;
          Line_Tail : Natural := Line'FIRST - 1;


          FUNCTION Get_Byte( Image : String )
              RETURN Compression_Instance.Object_Value IS
          BEGIN
              RETURN Compression_Instance.Object_Value'VALUE
                       ( "16#" & Image & "#" );
          EXCEPTION
              WHEN Standard.Constraint_Error =>
                  Error( "Could not parse byte: " & Image );
                  RETURN 0;
          END Get_Byte;


          PROCEDURE Process_Word( Word : IN String ) IS
              Head : CONSTANT Natural := Word'FIRST;
          BEGIN
              IF Set_Head_Now THEN
                  Set_Head_Now := False;
                  Get_Head := Start_At;
              END IF;
              List( Start_At ) := Get_Byte( Word( Head + 0 .. Head + 1 ) );
              Start_At := Start_At + 1;
              List( Start_At ) := Get_Byte( Word( Head + 2 .. Head + 3 ) );
              Start_At := Start_At + 1;
          EXCEPTION
              WHEN Standard.Constraint_Error =>
                  Error( "Constraint_Error Raised..." );
                  Error( "Could not parse word: " & Word );
                  Error( "Start => " & Image( Start_At ) );
                  Error( "First => " & Image( List'FIRST ) );
                  Error( "Last  => " & Image( List'LAST ) );
              WHEN Standard.Program_Error =>
                  Error( "PROGRAM_ERROR Raised???" );
                  Error( "Could not parse word: " & Word );
                  Error( "Start => " & Image( Start_At ) );
                  Error( "First => " & Image( List'FIRST ) );
                  Error( "Last  => " & Image( List'LAST ) );
          END Process_Word;


      BEGIN
          Next_Word( Line, Line_Head, Line_Tail );
          IF Line_Tail IN Line_Head .. Line_Tail THEN
              Start_At :=
                  Compression_Instance.Object_Index'VALUE
                    ( "16#" & Line( Line_Head .. Line_Tail ) & "#" ) * 2;
              LOOP
                  Line_Head := Line_Tail + 1;
                  Next_Word( Line, Line_Head, Line_Tail );
                  EXIT WHEN Line_Tail NOT IN Line_Head .. Line_Tail;
                  Process_Word( Line( Line_Head .. Line_Tail ) );
              END LOOP;
          END IF;
      EXCEPTION
          WHEN Standard.Constraint_Error =>
              Error( "COULD NOT PARSE: " & Line );
      END Process_Line;



      PROCEDURE Read_And_Process_Line IS
          Text : String( 1 .. 80 ) := ( OTHERS => ' ' );
          From : Natural := Text'FIRST;
          Most : Natural := 0;
          Skip : Boolean := False;
      BEGIN
          Read_Line
            ( From => Compression_Input_File,
              Text => Text,
              Last => Most,
              Head => From,
              Skip => Skip );
          IF NOT Skip THEN
              IF Compression_Parameters.Contains_Code_Separator
                   ( Text( From .. Most ) )
              THEN
                  Set_Head_Now := True;
              ELSE
                  Process_Line( Text( From .. Most ) );
              END IF;
          END IF;
      END Read_And_Process_Line;



  BEGIN
      Print( "Reading memory map file..." );
      WHILE NOT Text_IO.End_Of_File( File_State( Compression_Input_File ).File )
      LOOP
          Read_And_Process_Line;
      END LOOP;
      Close_File( Compression_Input_File );
      IF Start_At IN List'FIRST .. List'FIRST THEN
          List( Start_At ) := 0;
          Start_At := Start_At + 1;
          List( Start_At ) := 0;
      END IF;
      Last := Start_At - 1;
      Head := Get_Head;
      Dump( List( List'FIRST .. Start_At - 1 ), Get_Head, "Load:" );
  END Load_Memory;



  PROCEDURE Save_Memory
    ( List : IN Compression_Instance.Object_Array;
      Head : IN Compression_Instance.Object_Index;
      Last : IN Compression_Instance.Object_Index ) IS

      Next : Natural := 0;

      -- Convert word index to byte index...
      
      Starts : Compression_Instance.Object_Index :=
                   Compression_Instance.Object_Index'FIRST +
                   ( List'FIRST - Compression_Instance.Object_Index'FIRST ) * 2;

      Byte_1 : Compression_Instance.Object_Index := List'FIRST;
      Byte_2 : Compression_Instance.Object_Index := Byte_1;

      Buffer : String( 1 .. 80 );
      Length : Natural := Buffer'FIRST - 1;



      PROCEDURE Output( Text : IN String ) IS
      BEGIN
          Text_IO.Put_Line( File_State( Compression_Output_File ).File, Text );
      END Output;



      PROCEDURE Clear_Buffer IS
          Non_Zero_Found : Boolean := False;
      BEGIN
          IF Length > 0 THEN
              FOR Index IN Buffer'FIRST + 7 .. Length LOOP
                  Non_Zero_Found := Buffer( Index ) /= ' ' AND THEN
                                    Buffer( Index ) /= '0';
                  EXIT WHEN Non_Zero_Found;
              END LOOP;
              IF Non_Zero_Found THEN
                  Output( Buffer( Buffer'FIRST .. Length ) );
              END IF;
              Length := Buffer'FIRST - 1;
              Next := 0;
          END IF;
      END Clear_Buffer;



      FUNCTION Get_Byte( Byte : Compression_Instance.Object_Index )
          RETURN Compression_Instance.Object_Value IS
      BEGIN
          IF Byte IN List'FIRST .. Last THEN
              RETURN List( Byte );
          ELSE
              -- Create a full word by adding a special marker byte
              -- to the last odd byte.
              RETURN Parameters.Markers.List( Parameters.Markers.List'FIRST );
          END IF;
      END Get_Byte;



      PROCEDURE Add_Image( Image : String ) IS
      BEGIN
          Buffer( Length + 1 .. Length + Image'LENGTH ) := Image;
          Length := Length + Image'LENGTH;
      END Add_Image;



      PROCEDURE Add_Entry( Index : Compression_Instance.Object_Index ) IS
      BEGIN
          IF Length < Buffer'FIRST THEN
              Add_Image( Hex( Index ) & "   " );
          END IF;
          Add_Image( " " & Hex( Get_Byte( Byte_1 )) & Hex( Get_Byte( Byte_2 )));
          Next := Next + 1;
          IF Next = Maximum_Columns THEN
              Clear_Buffer;
          END IF;
      END Add_Entry;



  BEGIN
      Dump( List( List'FIRST .. Last ), Head, "Save:" );
      Print( "Writing memory map file..." );
      FOR Index IN Starts .. Starts + ( Last - List'FIRST ) / 2 LOOP
          Byte_2 := Compression_Instance.Object_Index'SUCC( Byte_1 );
          IF Head IN Byte_1 .. Byte_2 THEN
              Clear_Buffer;
              Output( Compression_Parameters.Fixed_Code_Separator );
          END IF;
          Add_Entry( Index );
          Byte_1 := Compression_Instance.Object_Index'SUCC( Byte_2 );
      END LOOP;
      Clear_Buffer;
      Close_File( Compression_Output_File );
  END Save_Memory;



END Compression_Support;
